using System;
using System.Collections.Generic;
using System.Drawing;
using System.IO;
using System.Numerics;
using System.Runtime.InteropServices;
using SlatedGameToolkit.Framework.Graphics.OpenGL;
using SlatedGameToolkit.Framework.Graphics.Render;
using SlatedGameToolkit.Framework.Graphics.Textures;
using SlatedGameToolkit.Framework.Graphics.Window;
using SlatedGameToolkit.Framework.Utilities;
using SlatedGameToolkit.Framework.Utilities.Collections;
using StbTrueTypeSharp;

namespace SlatedGameToolkit.Framework.Graphics.Text
{
    public class BitmapFont : IDisposable
    {
        private StbTrueType.stbtt_fontinfo info;
        private byte[] fontData;
        private bool disposed;
        private float scale;
        private int bitmapLength;
        private int pixelHeight;
        public int PixelHeight {
            get {
                return pixelHeight;
            }

            set {
                pixelHeight = value;
                scale = StbTrueType.stbtt_ScaleForPixelHeight(info, value);
            }
        }

        private float kerning;
        public float Kerning {
            get {
                return kerning;
            }

            set {
                kerning = value;
            }
        }
        private Dictionary<(float, char), int> bitmapLocations;
        private LRUCache<char, int> glyphIndexCache;
        private FontBitmap[] textureBuffers;
        private int drawingTo = 0;  

        GLContext glContext;

        /// <summary>
        /// Creates a font with the given data.
        /// </summary>
        /// <param name="fontData">The byte array containing all ttf data.</param>
        /// <param name="glContext">The OpenGL context to use to associate with.</param>
        /// <param name="cacheSize">The size of the caches.</param>
        /// <param name="bitmapLength">The length of one side of a bitmap, therefore, the total would be this squared.</param>
        /// <param name="numberOfTextures">The number of backing textures to use.</param>
        /// <param name="initialCharacters">An initial set of characters to load.</param>
        public unsafe BitmapFont(byte[] fontData, GLContext glContext = null, int cacheSize = 1024, int bitmapLength = 512, int numberOfTextures = 2, params char[] initialCharacters) {
            this.glContext = glContext ?? WindowContextsManager.CurrentGL;
            if (bitmapLength < 0) throw new ArgumentOutOfRangeException("bitmapLength");
            if (fontData == null || fontData.Length <= 0) throw new ArgumentException("fontData must be an array of bytes representing a TTF file.");
            glyphIndexCache = new LRUCache<char, int>(cacheSize);
            bitmapLocations = new Dictionary<(float, char), int>();
            this.fontData = fontData;

            textureBuffers = new FontBitmap[numberOfTextures];
            this.bitmapLength = bitmapLength;
            for (int i = 0; i < textureBuffers.Length; i++)
            {
                textureBuffers[i] = new FontBitmap(this.glContext, bitmapLength);
            }

            info = new StbTrueType.stbtt_fontinfo();

            fixed (byte* data = &fontData[0]) {
                StbTrueType.stbtt_InitFont(info, data, 0);
            }

            PixelHeight = 64;
        }

        public unsafe BitmapFont(string path, GLContext glContext = null, int cacheSize = 1024, int bitmapLength = 512, int numberOfTextures = 2, params char[] initialCharacters) :
        this(File.ReadAllBytes(path), glContext, cacheSize, bitmapLength, numberOfTextures, initialCharacters)
        {
        }

        /// <summary>
        /// Loads the requested character.
        /// </summary>
        /// <param name="character">The character to load.</param>
        /// <returns>True if already loaded or successfully loaded, false if could not load.</returns>
        public unsafe bool LoadGlyph(char character) {
            if (!(bitmapLocations.ContainsKey((scale, character)) && textureBuffers[bitmapLocations[(scale, character)]].Contains(scale, character))) {
                bitmapLocations.Remove((scale, character));
                int glyphIndex = glyphIndexCache.ComputeIfNonExistent(character, (c) => StbTrueType.stbtt_FindGlyphIndex(info, c));
                int x0 = 0, y0 = 0, x1 = 0, y1 = 0;
                StbTrueType.stbtt_GetGlyphBitmapBox(info, glyphIndex, scale, scale, &x0, &y0, &x1, &y1);
                if (!textureBuffers[drawingTo].CanAdd(x1 - x0, y1 - y0)) {
                    return false;
                }
                
                textureBuffers[drawingTo].Upload(character, x1 - x0, y1 - y0, scale, info, glyphIndex);
                bitmapLocations.Add((scale, character), drawingTo);
            }
            return true;
        }

        public unsafe void Draw(MeshBatch batch, double delta, float x, float y, string characters, Color color) {
            float horizontal = x;
            foreach (char c in characters)
            {
                int used = 0;
                while (!LoadGlyph(c)) {
                    used++;
                    drawingTo++;
                    if (drawingTo > textureBuffers.Length) {
                        drawingTo = 0;
                    }
                    if (used >= textureBuffers.Length) {
                        batch.End();
                        batch.Begin(Matrix4x4.Identity, delta);
                        used = 0;
                    }
                }
                ITexture texture = textureBuffers[bitmapLocations[(scale, c)]];
                RectangleF glyphBounds = textureBuffers[bitmapLocations[(scale, c)]].GetGlyphTextureBounds(scale, c);
                int width, height, vX, vY;
                glContext.GetViewport(out vX, out vY, out width, out height);
                float glyphWidth = glyphBounds.Width / width;
                float glyphHeight = glyphBounds.Height / height;
                RectangleMesh mesh = new RectangleMesh(new RectangleF(horizontal, y, glyphWidth, glyphHeight), glyphBounds.MultiplyBy(1f/bitmapLength), texture, color);
                horizontal += (glyphBounds.Width - kerning) / width;
                batch.Draw(mesh);
            }
        }

        public ITexture GetActiveTextureBacking() {
            return textureBuffers[drawingTo];
        }

        protected virtual void Dispose(bool disposing)
        {
            if (!disposed)
            {
                if (disposing)
                {
                }
                foreach (FontBitmap bitmap in textureBuffers)
                {
                    bitmap.Dispose();
                }
                disposed = true;
            }
        }

        ~BitmapFont()
        {
            Dispose(disposing: false);
        }

        public void Dispose()
        {
            Dispose(disposing: true);
            GC.SuppressFinalize(this);
        }
        private class FontBitmap : ITexture
        {
            private GLContext glContext;
            private bool disposed;
            private int bitmapLength;
            private int lastRow;
            private Point bitmapPosition;
            private Dictionary<(float, char), RectangleF> glyphBitmaps = new Dictionary<(float, char), RectangleF>();
            public uint Handle {get; private set;}

            public uint Width {get; private set;}

            public uint Height {get; private set;}

            public bool SingleChanneled => true;

            public unsafe FontBitmap(GLContext glContext, int length) {
                uint[] handles = new uint[1];
                this.glContext = glContext;
                glContext.GenTextures(handles.Length, handles);
                Handle = handles[0];
                glContext.BindTexture(TextureTarget.Texture2D, Handle);
                glContext.TexImage2D(TextureTarget.Texture2D, 0, InternalFormat.Red, (uint)length, (uint)length, 0, PixelFormat.Red, PixelType.UnsignedByte, IntPtr.Zero);
                glContext.TexParameteri(TextureTarget.Texture2D, TextureParameterName.TextureWrapS, (int)TextureWrapMode.ClampToEdge);
                glContext.TexParameteri(TextureTarget.Texture2D, TextureParameterName.TextureWrapT, (int)TextureWrapMode.ClampToEdge);
                glContext.TexParameteri(TextureTarget.Texture2D, TextureParameterName.TextureMinFilter, (int)TextureMinFilter.Nearest);
                glContext.TexParameteri(TextureTarget.Texture2D, TextureParameterName.TextureMagFilter, (int)TextureMagFilter.Linear);
                this.bitmapLength = length;
            }

            public bool Contains(float scale, char character) {
                return glyphBitmaps.ContainsKey((scale, character));
            }

            public RectangleF GetGlyphTextureBounds(float scale, char character) {
                return glyphBitmaps[(scale, character)];
            }

            public unsafe void Upload(char character, int width, int height, float scale, StbTrueType.stbtt_fontinfo info, int glyphIndex) {
                if (bitmapPosition.X + width < bitmapLength) { // Within the first row
                    if (bitmapPosition.Y - lastRow < height) {
                        bitmapPosition.Y = lastRow + height + 1;
                        if (bitmapPosition.Y >= bitmapLength) {
                            bitmapPosition = Point.Empty;
                            glyphBitmaps.Clear();
                        }
                    } // the prev row + the current bitmaps height will be the maximum height.
                    bitmapPosition.X += width + 1; // move the x position.
                    glContext.BindTexture(TextureTarget.Texture2D, Handle);
                    int w = 0, h = 0, xoff = 0, yoff  = 0;
                    byte* dataPtr = StbTrueType.stbtt_GetGlyphBitmap(info, scale, scale, glyphIndex, &w, &h, &xoff, &yoff);
                    IntPtr data = new IntPtr(dataPtr);
                    Rectangle drawLocation = new Rectangle(bitmapPosition.X - width, bitmapPosition.Y - height, width, height);
                    int prevAlignment = 0;
                    int[] prevAlignArr = new int[1];
                    glContext.GetIntegerv(GetPName.UnpackAlignment, prevAlignArr);
                    prevAlignment = prevAlignArr[0];
                    if (prevAlignment != 1) {
                        glContext.PixelStorei(PixelStoreParameter.UnpackAlignment, 1);
                    }
                    glContext.TexSubImage2D(TextureTarget.Texture2D, 0, drawLocation.X, drawLocation.Y, drawLocation.Width, drawLocation.Height, PixelFormat.Red, PixelType.UnsignedByte, data);
                    if (prevAlignment != 1) {
                        glContext.PixelStorei(PixelStoreParameter.UnpackAlignment, prevAlignment);
                    }
                    StbTrueType.stbtt_FreeBitmap(dataPtr, null);
                    glyphBitmaps.Add((scale, character), new Rectangle(drawLocation.X, bitmapLength - drawLocation.Y - drawLocation.Height, drawLocation.Width, drawLocation.Height));
                    
                } else { // Next row
                    lastRow = bitmapPosition.Y; // sets the previous row var to the previous row.
                    bitmapPosition.X = 0;
                    Upload(character, width, height, scale, info, glyphIndex);
                }
            }

            public bool CanAdd(int width, int height) {
                return (bitmapPosition.X + width < bitmapLength) && (bitmapPosition.Y + height < bitmapLength);
            }

            protected virtual void Dispose(bool disposing)
            {
                if (!disposed)
                {
                    if (disposing)
                    {
                    }
                    glContext.DeleteTextures(1, new uint[] {Handle});
                    disposed = true;
                }
            }

            ~FontBitmap()
            {
                Dispose(disposing: false);
            }

            public void Dispose()
            {
                Dispose(disposing: true);
                GC.SuppressFinalize(this);
            }
        }
    }
}